"""Affine Transformations"""

from typing import Callable
from typing import Literal
from typing import Sequence
from typing import Union

import numpy as np
from jax import nn
from jax import numpy as jnp
from jax.lax import conv_general_dilated as _conv_general_dilated
from jax.lax import conv_transpose as _conv_transpose
from jax.lax import ConvDimensionNumbers
from jaxtyping import Array
from jaxtyping import PyTree

from zephyr.building import initializers
from zephyr.building import template
from zephyr.building.template import validate
from zephyr.functools.partial import flexible
from zephyr.masking import apply_mask
from zephyr.project_typing import PaddingPreset
from zephyr.project_typing import Shape
from zephyr.project_typing import ShapeExpression


@flexible
def linear(
    params: PyTree,
    x: Array,
    out_dim: int,
    use_bias: bool = True,
    weights_initializer: initializers.Initializer = initializers.initializer_base,
    bias_initializer: initializers.Initializer = initializers.initializer_base,
    activation=lambda x: x,
) -> Array:
    validate(params["weights"], (x.shape[-1], out_dim), weights_initializer)
    z = x @ params["weights"]
    if use_bias:
        validate(params["bias"], (out_dim,), bias_initializer)
        z = z + params["bias"]
    z = activation(z)
    return z


@flexible
def branch_linear(
    params: PyTree,
    x: Array,
    num_branches: int,
    with_bias: bool = True,
    weights_initializer: initializers.Initializer = initializers.initializer_base,
    bias_initializer: initializers.Initializer = initializers.initializer_base,
    activation=lambda x: x,
) -> Array:
    """Branches the last dimension of `x` with each branch having the same dimension as the last dimension of `x`

    Example:
        if x.shape == (..., e) then after this function

            z = branch_linear (...x, num_branches...)
            z.shape == (..., num_branches, x.shape[-1])

    """
    validate(
        params,
        expression=lambda params: params["weights"].shape[-1] // x.shape[-1]
        == num_branches,
    )
    z = linear(
        params,
        x,
        x.shape[-1] * num_branches,
        with_bias,
        weights_initializer,
        bias_initializer,
        activation,
    )
    z = jnp.reshape(z, z.shape[:-1] + (num_branches, x.shape[-1]))

    return z


@flexible
def linear_like(
    params: PyTree,
    array_to_be_projected_to_desired_shape: Array,
    reference_array_with_desired_last_dimension: Array,
    use_bias: bool = True,
    weights_initializer: initializers.Initializer = initializers.initializer_base,
    bias_initializer: initializers.Initializer = initializers.initializer_base,
    activation=lambda x: x,
) -> Array:
    validate(
        params,
        expression=lambda params: reference_array_with_desired_last_dimension
        == jnp.ones(
            # array_to_be_projected_to_desired_shape.shape[:-1]
            # +
            (params["weights"].shape[-2],),
        ),
    )
    array_with_desired_shape = linear(
        params,
        array_to_be_projected_to_desired_shape,
        reference_array_with_desired_last_dimension.shape[-1],
        use_bias,
        weights_initializer,
        bias_initializer,
        activation,
    )

    return array_with_desired_shape


@flexible
def conv_general(
    params: PyTree,
    x: Array,
    num_spatial_axes: int,
    out_channels: int,
    kernel_shape: ShapeExpression,
    stride: ShapeExpression = 1,
    rate: ShapeExpression = 1,
    padding: Union[PaddingPreset, Sequence[tuple[int, int]]] = "SAME",
    dilation: ShapeExpression = 1,
    with_bias: bool = True,
    mask: None | Array = None,
    weights_initializer: initializers.Initializer = initializers.initializer_base,
    bias_initializer: initializers.Initializer = initializers.initializer_base,
    feature_group_count: int = 1,
    data_format: Literal["CHANNELS_LAST"] = "CHANNELS_LAST",
    activation=lambda x: x,
) -> Array:
    if data_format != "CHANNELS_LAST":
        raise NotImplementedError(f"`data_format` not supported")
    kernel_shape = _to_shape_tuple(kernel_shape, num_spatial_axes)
    padding = padding if type(padding) is str else _to_padding_tuple(padding)

    stride, kernel_dilation, lhs_dilation = map(
        lambda x: _to_shape_tuple(x, num_spatial_axes), [stride, rate, dilation]
    )

    dimension_numbers = _to_dimenstion_numbers(
        num_spatial_axes, data_format == "CHANNELS_LAST", False
    )
    num_axes = len(x.shape)

    uses_batches = num_axes != num_spatial_axes + 1
    if not uses_batches:
        x = jnp.expand_dims(x, axis=0)

    if x.shape[-1] % feature_group_count != 0:
        raise ValueError(
            f"Inputs channels: {x.shape[-1]} should be divisible by `feature_group_count`:{feature_group_count}"
        )

    weights_shape = kernel_shape + (x.shape[-1] // feature_group_count, out_channels)
    validate(params["weights"], weights_shape, weights_initializer)

    if mask is not None:
        params["weights"] = apply_mask(params["weights"], mask)

    z = _conv_general_dilated(
        x,
        params["weights"],
        window_strides=stride,
        padding=padding,
        lhs_dilation=lhs_dilation,
        rhs_dilation=kernel_dilation,
        dimension_numbers=dimension_numbers,
        feature_group_count=feature_group_count,
    )

    if with_bias:
        validate(params["bias"], (out_channels,), bias_initializer)
        bias = jnp.broadcast_to(params["bias"], z.shape)
        z = z + bias

    if not uses_batches:
        z = jnp.squeeze(z, 0)

    z = activation(z)

    return z


@flexible
def conv_1d(
    params: PyTree,
    x: Array,
    out_channels: int,
    kernel_shape: ShapeExpression,
    stride: ShapeExpression = 1,
    rate: ShapeExpression = 1,
    padding: Union[PaddingPreset, Sequence[tuple[int, int]]] = "SAME",
    dilation: ShapeExpression = 1,
    with_bias: bool = True,
    mask: None | Array = None,
    weights_initializer: initializers.Initializer = initializers.initializer_base,
    bias_initializer: initializers.Initializer = initializers.initializer_base,
    feature_group_count: int = 1,
    data_format: Literal["CHANNELS_LAST"] = "CHANNELS_LAST",
    activation=lambda x: x,
) -> Array:
    return conv_general(
        params,
        x,
        1,
        out_channels,
        kernel_shape,
        stride,
        rate,
        padding,
        dilation,
        with_bias,
        mask,
        weights_initializer,
        bias_initializer,
        feature_group_count,
        data_format,
        activation,
    )


@flexible
def conv_2d(
    params: PyTree,
    x: Array,
    out_channels: int,
    kernel_shape: ShapeExpression,
    stride: ShapeExpression = 1,
    rate: ShapeExpression = 1,
    padding: Union[PaddingPreset, Sequence[tuple[int, int]]] = "SAME",
    dilation: ShapeExpression = 1,
    with_bias: bool = True,
    mask: None | Array = None,
    weights_initializer: initializers.Initializer = initializers.initializer_base,
    bias_initializer: initializers.Initializer = initializers.initializer_base,
    feature_group_count: int = 1,
    data_format: Literal["CHANNELS_LAST"] = "CHANNELS_LAST",
    activation=lambda x: x,
) -> Array:
    return conv_general(
        params,
        x,
        2,
        out_channels,
        kernel_shape,
        stride,
        rate,
        padding,
        dilation,
        with_bias,
        mask,
        weights_initializer,
        bias_initializer,
        feature_group_count,
        data_format,
        activation,
    )


@flexible
def conv_3d(
    params: PyTree,
    x: Array,
    out_channels: int,
    kernel_shape: ShapeExpression,
    stride: ShapeExpression = 1,
    rate: ShapeExpression = 1,
    padding: Union[PaddingPreset, Sequence[tuple[int, int]]] = "SAME",
    dilation: ShapeExpression = 1,
    with_bias: bool = True,
    mask: None | Array = None,
    weights_initializer: initializers.Initializer = initializers.initializer_base,
    bias_initializer: initializers.Initializer = initializers.initializer_base,
    feature_group_count: int = 1,
    data_format: Literal["CHANNELS_LAST"] = "CHANNELS_LAST",
    activation=lambda x: x,
) -> Array:
    return conv_general(
        params,
        x,
        3,
        out_channels,
        kernel_shape,
        stride,
        rate,
        padding,
        dilation,
        with_bias,
        mask,
        weights_initializer,
        bias_initializer,
        feature_group_count,
        data_format,
        activation,
    )


@flexible
def conv_transpose_general(
    params: PyTree,
    x: Array,
    num_spatial_axes: int,
    out_channels: int,
    kernel_shape: ShapeExpression,
    stride: ShapeExpression = 1,
    padding: PaddingPreset | Sequence[tuple[int, int]] = "SAME",
    with_bias: bool = True,
    mask: Array | None = None,
    weights_initializer: initializers.Initializer = initializers.initializer_base,
    bias_initializer: initializers.Initializer = initializers.initializer_base,
    feature_group_count: int = 1,
    data_format: Literal["CHANNELS_LAST"] = "CHANNELS_LAST",
    activation=lambda x: x,
) -> Array:
    if data_format != "CHANNELS_LAST":
        raise NotImplementedError(f"`data_format` not supported")
    kernel_shape = _to_shape_tuple(kernel_shape, num_spatial_axes)
    padding = (
        padding
        if type(padding) is str
        else _to_padding_tuple(padding, num_spatial_axes)
    )

    stride = _to_shape_tuple(stride, num_spatial_axes)

    dimension_numbers = _to_dimenstion_numbers(
        num_spatial_axes, data_format == "CHANNELS_LAST", False
    )
    num_axes = len(x.shape)

    uses_batches = num_axes != num_spatial_axes + 1
    if not uses_batches:
        x = jnp.expand_dims(x, axis=0)

    if x.shape[-1] % feature_group_count != 0:
        raise ValueError(
            f"Inputs channels: {x.shape[-1]} should be divisible by `feature_group_count`:{feature_group_count}"
        )

    weights_shape = kernel_shape + (x.shape[-1] // feature_group_count, out_channels)
    validate(params["weights"], weights_shape, weights_initializer)

    if mask is not None:
        params["weights"] = apply_mask(params["weights"], mask)

    z = _conv_transpose(
        x,
        params["weights"],
        strides=stride,
        padding=padding,
        dimension_numbers=dimension_numbers,
    )

    if with_bias:
        validate(params["bias"], (out_channels,), bias_initializer)
        bias = jnp.broadcast_to(params["bias"], z.shape)
        z = z + bias

    if not uses_batches:
        z = jnp.squeeze(z, 0)

    z = activation(z)

    return z


@flexible
def conv_transpose_1d(
    params: PyTree,
    x: Array,
    out_channels: int,
    kernel_shape: ShapeExpression,
    stride: ShapeExpression = 1,
    padding: PaddingPreset | Sequence[tuple[int, int]] = "SAME",
    with_bias: bool = True,
    mask: Array | None = None,
    weights_initializer: initializers.Initializer = initializers.initializer_base,
    bias_initializer: initializers.Initializer = initializers.initializer_base,
    feature_group_count: int = 1,
    data_format: Literal["CHANNELS_LAST"] = "CHANNELS_LAST",
    activation=lambda x: x,
) -> Array:
    return conv_transpose_general(
        params,
        x,
        1,
        out_channels,
        kernel_shape,
        stride,
        padding,
        with_bias,
        mask,
        weights_initializer,
        bias_initializer,
        feature_group_count,
        data_format,
        activation,
    )


@flexible
def conv_transpose_2d(
    params: PyTree,
    x: Array,
    out_channels: int,
    kernel_shape: ShapeExpression,
    stride: ShapeExpression = 1,
    padding: PaddingPreset | Sequence[tuple[int, int]] = "SAME",
    with_bias: bool = True,
    mask: Array | None = None,
    weights_initializer: initializers.Initializer = initializers.initializer_base,
    bias_initializer: initializers.Initializer = initializers.initializer_base,
    feature_group_count: int = 1,
    data_format: Literal["CHANNELS_LAST"] = "CHANNELS_LAST",
    activation=lambda x: x,
) -> Array:
    return conv_transpose_general(
        params,
        x,
        2,
        out_channels,
        kernel_shape,
        stride,
        padding,
        with_bias,
        mask,
        weights_initializer,
        bias_initializer,
        feature_group_count,
        data_format,
        activation,
    )


@flexible
def conv_transpose_3d(
    params: PyTree,
    x: Array,
    out_channels: int,
    kernel_shape: ShapeExpression,
    stride: ShapeExpression,
    padding: PaddingPreset | Sequence[tuple[int, int]] = "SAME",
    with_bias: bool = True,
    mask: Array | None = None,
    weights_initializer: initializers.Initializer = initializers.initializer_base,
    bias_initializer: initializers.Initializer = initializers.initializer_base,
    feature_group_count: int = 1,
    data_format: Literal["CHANNELS_LAST"] = "CHANNELS_LAST",
    activation=lambda x: x,
) -> Array:
    return conv_transpose_general(
        params,
        x,
        3,
        out_channels,
        kernel_shape,
        stride,
        padding,
        with_bias,
        mask,
        weights_initializer,
        bias_initializer,
        feature_group_count,
        data_format,
        activation,
    )


def _to_dimenstion_numbers(
    num_spatial_axes: int, channels_last: bool, transpose: bool
) -> ConvDimensionNumbers:
    num_axes = num_spatial_axes + 2
    if channels_last:
        spatial_axes = tuple(range(1, num_axes - 1))
        image_dimension_numbers = (0, num_axes - 1) + spatial_axes
    else:
        spatial_axes = tuple(range(2, num_axes))
        image_dimension_numbers = (0, 1) + spatial_axes

    if transpose:
        kernel_dimension_numbers = (num_axes - 2, num_axes - 1) + tuple(
            range(num_axes - 2)
        )
    else:
        kernel_dimension_numbers = (num_axes - 1, num_axes - 2) + tuple(
            range(num_axes - 2)
        )

    return ConvDimensionNumbers(
        lhs_spec=image_dimension_numbers,
        rhs_spec=kernel_dimension_numbers,
        out_spec=image_dimension_numbers,
    )


def _to_padding_tuple(
    padding: tuple[int, int], num_spatial_axes: int
) -> Sequence[tuple[int, int]] | tuple[tuple[int, int]]:
    if type(padding[0]) is int:
        return (padding,) * num_spatial_axes
    elif len(padding) == 1:
        return tuple(padding) * num_spatial_axes
    raise ValueError(
        f"Padding {padding} must of length {num_spatial_axes} but got {len(padding)}"
    )


def _to_shape_tuple(x: ShapeExpression, num_times: int) -> Sequence[int]:
    if type(x) is int:
        return (x,) * num_times
    if len(x) == 1:
        return tuple(x * num_times)
    if len(x) == num_times:
        return tuple(x)
    raise ValueError(f"Did not understand {x}")



from zephyr.building._template.array import array
from zephyr.building._template.array import array_equal
from zephyr.building._template.validation import validate

# from functools import partial
# from typing import Callable
# from typing import Tuple
# from typing import Union

# from jax import random
# from jaxtyping import Array
# from jaxtyping import PyTree

# from zephyr.building.initializers import Initializer
# from zephyr.building.initializers import initializer_base
# from zephyr.building.skeleton import Skeleton
# from zephyr.project_typing import ArrayTemplate
# from zephyr.project_typing import Shape


# def array(shape: Shape, initializer: Initializer = initializer_base) -> ArrayTemplate:
#     return partial(initializer, shape=shape)


# def validate(
#     params: Union[Array, Skeleton],
#     shape: Shape,
#     initializer: Initializer = initializer_base,
# ) -> None:
#     if type(params) is Skeleton:
#         params == array(shape, initializer)
#     else:
#         if params.shape != shape:
#             raise ValueError(
#                 f"Incompatible shapes: shape of `params`: {params.shape} should be {shape} as specified."
#             )



from jax import nn
from jax import numpy as jnp
from jaxtyping import Array
from jaxtyping import PyTree

from zephyr._nets.linear import branch_linear
from zephyr._nets.linear import linear
from zephyr._nets.linear import linear_like
from zephyr._nets.mlp import mlp
from zephyr._nets.norm import layer_norm
from zephyr.building import initializers
from zephyr.building.template import validate
from zephyr.functools.partial import flexible


def fmap(x: Array) -> Array:
    return nn.elu(x) + 1


@flexible
def linear_attention(queries: Array, keys: Array, values: Array) -> Array:
    q = fmap(queries)
    k = fmap(keys)

    keys_transposed = jnp.moveaxis(k, -2, -1)
    scores = q @ (keys_transposed @ values)
    normalizer = q @ (keys_transposed @ jnp.ones_like(values))

    answers = scores / (normalizer + 1e-17)
    return answers


@flexible
def multi_head_linear_attention(
    queries, keys, values, num_heads: int, activation=lambda x: x
) -> Array:
    new_shape = queries.shape[:-1] + (num_heads, -1)
    queries = jnp.reshape(queries, new_shape)
    keys = jnp.reshape(queries, new_shape)
    values = jnp.reshape(queries, new_shape)

    # queries, keys, values [..., s, h, h//e]
    #                       [...,-3,-2,-1]

    queries = jnp.moveaxis(queries, -2, -3)
    keys = jnp.moveaxis(keys, -2, -3)
    values = jnp.moveaxis(values, -2, -3)

    multi_head_answers = linear_attention(queries, keys, values)  # [..., h, s, e]

    multi_head_answers = jnp.moveaxis(multi_head_answers, -2, -3)  # [..., s , h, h//e]

    combined_heads = jnp.reshape(
        multi_head_answers, multi_head_answers.shape[:-2] + (-1,)
    )

    combined_heads = activation(combined_heads)

    return combined_heads


@flexible
def linear_transformer_block(
    params: PyTree,
    queries: Array,
    keys: Array,
    values: Array,
    num_heads: int,
    embed_dim: int,
    mlp_dim: int,
    with_bias: bool = True,
    weights_initializer: initializers.Initializer = initializers.initializer_base,
    bias_initializer: initializers.Initializer = initializers.initializer_base,
    activation=lambda x: x,
) -> Array:
    queries = linear(
        params["linear"]["queries"],
        queries,
        embed_dim,
        with_bias,
        weights_initializer,
        bias_initializer,
    )
    keys = linear(
        params["linear"]["keys"],
        keys,
        embed_dim,
        with_bias,
        weights_initializer,
        bias_initializer,
    )
    values = linear(
        params["linear"]["values"],
        values,
        embed_dim,
        with_bias,
        weights_initializer,
        bias_initializer,
    )

    # values don't have to be projected since its embedding dimension was the reference
    z = multi_head_linear_attention(
        queries,
        keys,
        values,
        num_heads,
    )

    z = layer_norm(params["layer_norm"][0], z + queries, -1, True, True)
    z = layer_norm(
        params["layer_norm"][1],
        z
        + mlp(
            params["mlp"],
            z,
            [mlp_dim, embed_dim],
            weight_initializer=weights_initializer,
            bias_initializer=bias_initializer,
        ),
        -1,
        True,
        True,
    )

    z = activation(z)

    return z



from jax import random

from zephyr.functools.partial import flexible
from zephyr.masking import apply_mask
from zephyr.project_typing import Array
from zephyr.project_typing import KeyArray


@flexible
def dropout(_params, key: KeyArray, x: Array, drop_probability: float) -> Array:
    mask = 1.0 * (random.uniform(key, x.shape) > drop_probability)
    keep_probability = 1 - drop_probability
    scale_factor_so_that_the_sums_in_the_next_layers_are_in_the_same_range = 1 / (
        keep_probability + 1e-19
    )
    return (
        apply_mask(x, mask)
        * scale_factor_so_that_the_sums_in_the_next_layers_are_in_the_same_range
    )



from typing import Callable
from typing import Literal
from typing import Tuple

import numpy as np
from jax import numpy as jnp
from jax import random
from jaxtyping import Array

from zephyr.project_typing import KeyArray
from zephyr.project_typing import Shape

Initializer = Callable[[KeyArray, Tuple[int, ...]], Array]
from typing import Union
from jax.nn.initializers import truncated_normal


def initializer_base(
    key: KeyArray, shape: Shape, feature_shape: int | None = None
) -> Array:
    feature_shape = feature_shape or shape[0]  # shape[0] is assumed to be feature_shape
    standard_deviation = 1 / np.sqrt(feature_shape)

    if len(shape) > 1:
        return truncated_normal(standard_deviation)(key, shape)

    return zeros(key, shape)


def ones(key: KeyArray, shape: Shape) -> Array:
    return jnp.ones(shape)


def zeros(key: KeyArray, shape: Shape) -> Array:
    return jnp.zeros(shape)


def uniform(key: KeyArray, shape: Shape) -> Array:
    return random.uniform(key, shape)


def normal(key: KeyArray, shape: Shape) -> Array:
    return random.normal(key, shape)


def normal_scaled_by_rsqrt(key: KeyArray, shape: Shape):
    embed_dim = shape[-1]
    scale = 1 / (embed_dim**0.5)
    return scale * normal(key, shape)



from typing import Callable
from typing import Literal
from typing import NewType
from typing import Sequence
from typing import Union

from jaxtyping import Array

# KeyArray = NewType("KeyArray", Array)

KeyArray = Array
Shape = tuple[int, ...]
ActivationFunctionsWithKnownGain = Literal[
    "linear",
    "conv",  # corresponds to any conv (1d,2d, 1d transpose, etc)
    "sigmoid",
    "tanh",
    "relu",
    "leaky_relu",
]

PaddingPreset = Literal["SAME", "VALID", "CIRCULAR", "CAUSAL"]


ArrayTemplate = Callable[[KeyArray], Array]

ShapeExpression = Union[int, Sequence[int]]



from typing import Sequence
from typing import Union

import numpy as np
from jax import lax
from jax import numpy as jnp
from jax.lax import reduce_window
from jaxtyping import Array

from zephyr.functools.partial import flexible
from zephyr.project_typing import PaddingPreset
from zephyr.project_typing import ShapeExpression


@flexible
def max_pool(
    _params,
    x: Array,
    window_shape: ShapeExpression,
    stride: ShapeExpression = 1,
    padding: PaddingPreset = "SAME",
    channel_axis: int = -1,
    activation=lambda x: x,
) -> Array:
    if channel_axis != -1 and len(x.shape) != channel_axis:
        raise NotImplementedError("only channel last implemented")

    window_shape, strides = map(
        lambda s: _to_shape_tuple(x, s, channel_axis), [window_shape, stride]
    )

    return reduce_window(x, -jnp.inf, lax.max, window_shape, strides, padding)


@flexible
def avg_pool(
    _params,
    x: Array,
    window_shape: ShapeExpression,
    stride: ShapeExpression = 1,
    padding: PaddingPreset = "SAME",
    channel_axis: int = -1,
    activation=lambda x: x,
) -> Array:
    if channel_axis != -1 and len(x.shape) != channel_axis:
        raise NotImplementedError("only channel last implemented")

    window_shape, strides = map(
        lambda s: _to_shape_tuple(x, s, channel_axis), [window_shape, stride]
    )

    sum_pool = reduce_window(x, 0.0, lax.add, window_shape, strides, padding)
    if padding == "VALID":
        return sum_pool / np.prod(window_shape)

    window_counter_shape = [
        (v if w != 1 else 1) for (v, w) in zip(x.shape, window_shape)
    ]
    window_counts = lax.reduce_window(
        jnp.ones(window_counter_shape, x.dtype),
        0.0,
        lax.add,
        window_shape,
        strides,
        padding,
    )
    return sum_pool / window_counts


def _to_shape_tuple(
    x: Array,
    s: ShapeExpression,
    channel_axis: int = -1,
) -> Sequence[int]:
    if type(s) is int:
        if channel_axis < 0:
            return (1,) + tuple(
                s if d != channel_axis else 1 for d in range(1, len(x.shape))
            )
    elif len(s) < len(x.shape):
        return (1,) * (len(x.shape) - len(s)) + tuple(s)
    if len(x.shape) == len(s):
        return tuple(s)
    raise ValueError(f"Did not understand the inputs")



from collections.abc import Sequence
from functools import partial
from typing import Any
from warnings import warn

import jax
import numpy as np
from jax import Array
from jax import numpy as jnp
from jax import random

from zephyr.building._template.array import array_equal
from zephyr.project_typing import ArrayTemplate
from zephyr.project_typing import KeyArray
from zephyr.project_typing import Shape

# match jax._src.basearray typing
Shard = Any
Sharding = Any
Device = Any


class Skeleton(Array):
    def __init__(self, key: KeyArray):
        self._contents = {}
        self._key = key

    def materialize(self):
        """This initializes the arrays at the leaves using the appropriate initializers"""
        if type(self._contents) is dict:
            d = {}
            for k in self._contents:
                r = self._contents[k]
                if callable(r):
                    self._key, key = random.split(self._key)
                    r = r(key)
                else:  # is an skeletal
                    r = r.materialize()
                d[k] = r
            return d
        else:  # array_template
            self._key, key = random.split(self._key)
            return self._contents(key)  # array

    def __jax_array__(self):
        return self.materialize()

    def __array__(self):
        return self.materialize()

    def __getitem__(self, key):
        if key in self._contents:
            return self._contents[key]
        else:
            self._key, new_key = random.split(self._key)
            new_params = Skeleton(new_key)
            self._contents[key] = new_params
            return self._contents[key]

    def _has_been_associated_with_an_array_template_before(self) -> bool:
        return type(self._contents) is partial

    def __eq__(self, array_template: ArrayTemplate) -> bool:
        # todo: note to self <- __eq__ might be replaced with something else, as validate becomes more developed
        if self._has_been_associated_with_an_array_template_before():
            if array_equal(self._contents, array_template):
                warn(
                    f"Warning: params has been set before with shape {self._contents.keywords['shape']} "
                    f"and being set now with a SAME shape {array_template.keywords['shape']}.\n\n"
                    "Please make sure that this is intentional.",
                    RuntimeWarning,
                )  # still unsure, what to do.. what if the user just wants to validate twice ?
                return True

            raise ValueError(
                f"params has been set before with shape {self._contents.keywords['shape']} and being set now with the DIFFERENT shape {array_template.keywords['shape']}"
            )  # this is definitely an error
        self._contents = array_template
        return True

    def __add__(self, x):
        return self.materialize() + x

    def __radd__(self, x):
        return x + self.materialize()

    def __sub__(self, x):
        return self.materialize() - x

    def __rsub__(self, x):
        return x - self.materialize()

    def __mul__(self, x):
        return self.materialize() * x

    def __rmul__(self, x):
        return x * self.materialize()

    def __matmul__(self, x):
        return self.materialize() @ x

    def __rmatmul__(self, x):
        return x @ self.materialize()

    def __truediv__(self, x):
        return self.materialize() / x

    def __rtruediv__(self, x):
        return x / self.materialize()

    def __pow__(self, n):
        return self.materialize() ** n

    def __neg__(self):
        return -self.materialize()

    @property
    def dtype(self) -> np.dtype:
        return np.float32

    @property
    def ndim(self) -> int:
        return len(self.materialize().shape)

    @property
    def size(self) -> int:
        return np.prod(self.materialize().shape)

    @property
    def shape(self) -> Shape:
        return self.materialize().shape

    def addressable_data(self, index: int) -> Array:
        return self.materialize().addressable_data(index)

    @property
    def addressable_shards(self) -> Sequence[Shard]:
        return self.materialize().addressable_shards

    @property
    def global_shards(self) -> Sequence[Shard]:
        return self.materialize().global_shards

    @property
    def is_fully_addressable(self) -> bool:
        return self.materialize().is_fully_addressable

    @property
    def is_fully_replicated(self) -> bool:
        return self.materialize().is_fully_replicated

    @property
    def sharding(self) -> Sharding:
        return self.materialize().sharding

    @property
    def committed(self) -> bool:
        return (
            self.materialize().commited
        )  # this might return an AttributeError, but that comes from jax initializing an array without this attribute <- not zephyr problem (so far)

    @property
    def device(self) -> Device | Sharding:
        return jax.default_device()

    def copy_to_host_async(self) -> None:
        return



from functools import partial

from zephyr.building.initializers import Initializer
from zephyr.building.initializers import initializer_base
from zephyr.project_typing import ArrayTemplate
from zephyr.project_typing import Shape

ArrayTemplate_ = partial[ArrayTemplate]


def array(shape: Shape, initializer: Initializer = initializer_base) -> ArrayTemplate:
    return partial(initializer, shape=shape)


def array_equal(array_1: ArrayTemplate_, array_2: ArrayTemplate_) -> bool:
    return (
        array_1.func == array_2.func
        and array_1.keywords["shape"] == array_2.keywords["shape"]
    )



from typing import Literal
from typing import Optional

import numpy as np
from jax import lax
from jax import nn
from jax import numpy as jnp
from jaxtyping import Array
from jaxtyping import PyTree

from zephyr._nets.linear import branch_linear
from zephyr._nets.linear import linear
from zephyr.building import initializers
from zephyr.building.template import validate
from zephyr.functools.partial import flexible
from zephyr.masking import apply_attention_mask


@flexible
def single_head_attention(
    params: PyTree,
    queries: Array,
    keys: Array,
    values: Array,
    masks: Optional[Array] = None,
    with_bias: bool = True,
    weights_initializer: initializers.Initializer = initializers.initializer_base,
    bias_initializer: initializers.Initializer = initializers.initializer_base,
    activation=lambda x: x,
) -> Array:
    keys = linear(
        params["linear_keys"],
        keys,
        keys.shape[-1],
        with_bias,
        weights_initializer,
        bias_initializer,
    )
    queries = linear(
        params["linear_queries"],
        queries,
        keys.shape[-1],
        with_bias,
        weights_initializer,
        bias_initializer,
    )
    values = linear(
        params["linear_values"],
        values,
        values.shape[-1],
        with_bias,
        weights_initializer,
        bias_initializer,
    )

    # keys [... s k]
    # queries [... p k]
    # values [... s v]
    # target [... p v]

    scores = queries @ jnp.moveaxis(keys, -1, -2) / np.sqrt(keys.shape[-1])
    if masks is not None:
        scores = apply_attention_mask(scores, masks)
    attention_map = nn.softmax(scores, axis=-1)

    answers = attention_map @ values
    answers = activation(answers)
    return answers


@flexible
def multi_head_attention(
    params: PyTree,
    queries: Array,  # [batch, seq_q, model_dim]
    keys: Array,  # [batch, seq_k, model_dim]
    values: Array,  # [batch, seq_k, model_dim]
    num_heads: int,
    masks: Optional[Array] = None,
    is_causal: bool = False,
    implementation: Literal["xla", "cudnn"] = "cudnn",
    with_bias: bool = True,
    weights_initializer: initializers.Initializer = initializers.initializer_base,
    bias_initializer: initializers.Initializer = initializers.initializer_base,
) -> Array:
    model_dim = queries.shape[-1]
    head_dim = model_dim // num_heads
    assert model_dim % num_heads == 0, "model_dim must be divisible by num_heads"

    Q = linear(
        params["q_proj"],
        queries,
        num_heads * head_dim,
        with_bias,
        weights_initializer,
        bias_initializer,
    )
    K = linear(
        params["k_proj"],
        keys,
        num_heads * head_dim,
        with_bias,
        weights_initializer,
        bias_initializer,
    )
    V = linear(
        params["v_proj"],
        values,
        num_heads * head_dim,
        with_bias,
        weights_initializer,
        bias_initializer,
    )

    def split_heads(x: Array) -> Array:
        return jnp.reshape(
            x, x.shape[:-1] + (num_heads, head_dim)
        )  # [..., heads, head_dim]

    # Q = jnp.moveaxis(split_heads(Q), -2, 1)  # [batch, heads, seq_q, head_dim]
    # K = jnp.moveaxis(split_heads(K), -2, 1)  # [batch, heads, seq_k, head_dim]
    # V = jnp.moveaxis(split_heads(V), -2, 1)  # [batch, heads, seq_k, head_dim]

    if implementation == "cudnn":
        Q = lax.convert_element_type(Q, jnp.bfloat16)  # TODO make this flexible
        K = lax.convert_element_type(K, jnp.bfloat16)  # TODO make this flexible
        V = lax.convert_element_type(V, jnp.bfloat16)  # TODO make this flexible

    attn_output = nn.dot_product_attention(
        Q, K, V, mask=masks, is_causal=is_causal, implementation=implementation
    )

    # attn_output = jnp.moveaxis(attn_output, 1, -2)  # [batch, seq_q, heads, head_dim]
    attn_output = jnp.reshape(attn_output, attn_output.shape[:-2] + (model_dim,))

    output = linear(
        params["out_proj"],
        attn_output,
        model_dim,
        with_bias,
        weights_initializer,
        bias_initializer,
    )

    return output


@flexible
def multi_head_self_attention(
    params: PyTree,
    x: Array,  # [batch, seq_k, model_dim]
    num_heads: int,
    masks: Optional[Array] = None,
    is_causal: bool = False,
    implementation: Literal["xla", "cudnn"] = "cudnn",
    with_bias: bool = True,
    weights_initializer: initializers.Initializer = initializers.initializer_base,
    bias_initializer: initializers.Initializer = initializers.initializer_base,
) -> Array:
    return multi_head_attention(
        params,
        x,
        x,
        x,
        num_heads,
        masks,
        is_causal,
        implementation,
        with_bias,
        weights_initializer,
        bias_initializer,
    )



from zephyr._nets.attention import multi_head_attention
from zephyr._nets.attention import multi_head_self_attention
from zephyr._nets.attention import single_head_attention
from zephyr._nets.composition import sequential
from zephyr._nets.dropout import dropout
from zephyr._nets.embed import token_embed
from zephyr._nets.linear import branch_linear
from zephyr._nets.linear import conv_1d
from zephyr._nets.linear import conv_2d
from zephyr._nets.linear import conv_3d
from zephyr._nets.linear import conv_general
from zephyr._nets.linear import conv_transpose_1d
from zephyr._nets.linear import conv_transpose_2d
from zephyr._nets.linear import conv_transpose_3d
from zephyr._nets.linear import conv_transpose_general
from zephyr._nets.linear import linear
from zephyr._nets.linear import linear_like
from zephyr._nets.linear_attention import linear_transformer_block
from zephyr._nets.linear_attention import multi_head_linear_attention
from zephyr._nets.mlp import mlp
from zephyr._nets.norm import layer_norm
from zephyr._nets.pool import avg_pool
from zephyr._nets.pool import max_pool
from zephyr.functools.composition import chain
from zephyr.functools.composition import skip



from typing import Callable

from jax import nn
from jax import numpy as jnp
from jaxtyping import Array
from jaxtyping import PyTree

from zephyr._nets.linear import linear
from zephyr.building import initializers
from zephyr.building import template
from zephyr.building.template import validate
from zephyr.functools.partial import flexible


@flexible
def mlp(
    params: PyTree,
    x: Array,
    out_dims: list[int],
    activation=nn.relu,
    activation_final=lambda x: x,
    use_bias: bool = True,
    weight_initializer: initializers.Initializer = initializers.initializer_base,
    bias_initializer: initializers.Initializer = initializers.initializer_base,
):
    validate(
        params,
        expression=lambda params: [params[i]["weights"].shape[-2] for i in params]
        == out_dims,
    )
    for i, target_out in enumerate(out_dims[:-1]):
        x = activation(
            linear(
                params[i], x, target_out, use_bias, weight_initializer, bias_initializer
            )
        )

    if len(out_dims[:-1]) == 0:
        i = 0
    else:
        i += 1
    x = linear(
        params[i], x, out_dims[i], use_bias, weight_initializer, bias_initializer
    )

    x = activation_final(x)

    return x



from typing import Optional

from jax import nn
from jax import numpy as jnp
from jax.lax import rsqrt
from jaxtyping import Array
from jaxtyping import PyTree

from zephyr._nets.linear import branch_linear
from zephyr._nets.linear import linear
from zephyr.building import initializers
from zephyr.building import template
from zephyr.building.initializers import Initializer
from zephyr.building.template import validate
from zephyr.functools.partial import hole_aware
from zephyr.masking import apply_attention_mask


@hole_aware
def layer_norm(
    params: PyTree,
    x: Array,
    axis: int,
    create_scale: bool = True,
    create_offset: bool = True,
    eps: float = 1e-16,
    scale_initializer: Initializer = initializers.ones,
    offset_initializer: Initializer = initializers.zeros,
    activation=lambda x: x,
) -> Array:
    mean = jnp.mean(x, axis=axis, keepdims=True)
    variance = jnp.var(x, axis=axis, keepdims=True)

    axis = _relative_axis_to_absolute_axis(axis, x)
    shape = tuple(x.shape[axis] if i == axis else 1 for i, d in enumerate(x.shape))

    scale = jnp.array([1.0])
    if create_scale:
        validate(params["scale"], shape, scale_initializer)
        scale = params["scale"]
    scale = jnp.broadcast_to(scale, x.shape)

    offset = jnp.array([0.0])
    if create_offset:
        validate(params["offset"], shape, offset_initializer)
        offset = params["offset"]
    offset = jnp.broadcast_to(offset, x.shape)

    inversion = scale * rsqrt(variance + eps)
    normalized = inversion * (x - mean) + offset

    normalized = activation(normalized)

    return normalized


def _relative_axis_to_absolute_axis(axis: int, reference_array: Array) -> int:
    return tuple(range(reference_array.ndim))[axis]



"""TODO: Decide if these should go in _nets: they are increasingly useful in creation of nets"""

from functools import partial
from functools import wraps
from typing import Any
from typing import Callable
from typing import Sequence

from jax import random
from jaxtyping import Array

from zephyr.functools.partial import flexible


# @flexible
def thread(
    functions: Sequence[Callable], t: Any, split_rule: Callable = lambda x, i: x
) -> Sequence[Callable]:
    threaded_functions = []
    i = -1
    for i, fn in enumerate(functions[:-1]):
        t, t_sub = split_rule(t, i)
        threaded_functions.append(partial(fn, t_sub))

    threaded_functions.append(partial(functions[-1], t))

    return threaded_functions


@flexible
def chain(functions: Sequence[Callable]) -> Callable:
    @flexible
    def f(x):
        for fn in functions:
            x = fn(x)

        return x

    return f


@flexible
def params_split(params, i):
    return params[i], params[i + 1]


@flexible
def key_split(key, i):
    return random.split(key)


@flexible
def identity_split(x, i):
    return x


@flexible
def skip(f):
    @flexible
    @wraps(f)
    def inner(x):
        return x + f(x)

    return inner


thread_params = partial(thread, split_rule=params_split)
thread_key = partial(thread, split_rule=key_split)
thread_identity = partial(thread, split_rule=identity_split)



from typing import Any

from jaxtyping import Array
from jaxtyping import PyTree

from zephyr.functools.partial import flexible

Leaf = Any
Tag = Any


@flexible
def get_immediate_tags(params: dict, context=[]):
    if isinstance(params, Array):
        return context
    return {k: get_immediate_tags(v, context=[k]) for k, v in params.items()}


@flexible
def get_lineage_tags(params: dict, context=[]):
    if isinstance(params, Array):
        return context
    return {k: get_lineage_tags(v, context=context + [k]) for k, v in params.items()}



from typing import Optional

from jax import numpy as jnp
from jaxtyping import Array
from jaxtyping import PyTree

from zephyr.building._template.validation import validate
from zephyr.building.initializers import Initializer
from zephyr.building.initializers import normal_scaled_by_rsqrt
from zephyr.functools.partial import deriving_holes
from zephyr.functools.partial import hole_aware


@hole_aware
@deriving_holes
def token_embed(
    params: PyTree,
    x_token_ids: Array,
    vocab_size: int,
    embed_dim: int,
    initial_embedding_matrix: Optional[Array] = None,
    initializer: Initializer = normal_scaled_by_rsqrt,
    activation=lambda x: x,
) -> Array:
    if initial_embedding_matrix is not None:
        validate(
            params["token_embeddings"],
            (vocab_size, embed_dim),
            initializer=lambda key, shape: initial_embedding_matrix,
        )
        validate(
            params,
            expression=lambda params: params["token_embeddings"].shape
            == (vocab_size, embed_dim),
        )
    else:
        validate(params["token_embeddings"], (vocab_size, embed_dim), initializer)
    z = jnp.asarray(params["token_embeddings"])[(x_token_ids,)]
    z = activation(z)
    return z



# ?
# todo: maybe a sequential utility?
# what are good ways to compose functions relevant for DNNs?
from functools import partial
from functools import wraps
from typing import Callable
from typing import Sequence
from typing import Union

from jaxtyping import Array
from jaxtyping import PyTree

from zephyr.functools import composition
from zephyr.functools.composition import chain
from zephyr.functools.composition import thread_params
from zephyr.functools.partial import hole_aware
from zephyr.functools.partial import placeholder_hole as _

Params = PyTree
Layer = Callable[[Params, Array], Array]


@hole_aware
def sequential(
    params: PyTree, x: Array, layers: Sequence[Callable[[Params, Array], Array]]
) -> Array:
    return chain(thread_params(layers, params))(x)



"""zephyr
A neural network library on top of JAX allowing for easy and fast neural network designing, creation, and manipulation

Early Stage
"""

from zephyr.building.tracing import trace

__version__ = "0.0.21"
__author__ = "Marko Zolo Gozano Untalan"



from jax import numpy as jnp
from jaxtyping import Array


def apply_mask(x: Array, masks: Array) -> Array:
    return x * masks


def apply_attention_mask(x: Array, masks: Array) -> Array:
    # put large negative values instead of zeros
    return apply_mask(x, masks) + apply_mask((-1e16) * jnp.ones_like(x), 1 - masks)



from setuptools import find_packages
from setuptools import setup

setup(
    name="z-zephyr",
    version="0.0.21",
    description="Zephyr is a declarative neural network library on top of JAX allowing for easy and fast neural network designing, creation, and manipulation",
    url="https://github.com/mzguntalan/zephyr",
    author="Marko Zolo Gozano Untalan",
    author_email="mzguntalan@gmail.com",
    license="Apache-2.0",
    packages=find_packages(),
    install_requires=["jax>=0.7.1", "jaxlib>=0.7.1"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: GPU",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
    ],
)



from typing import Callable
from typing import Tuple
from typing import Union

from jaxtyping import Array
from jaxtyping import PyTree

from zephyr.building.skeleton import Skeleton as Tracer
from zephyr.project_typing import KeyArray


def trace(f: Callable, key: KeyArray, *args) -> PyTree:
    params = Tracer(key)
    f(params, *args)

    return params.materialize()



"""Module that would be possibly helpful for partial
application of functions in python


Notes:
    - experimental: might or might not be useful for the 
    rest of the lib
    - hopefully: it helps create more readable code
"""

from functools import wraps
from itertools import chain
from itertools import repeat
from typing import Any
from typing import Callable
from typing import ParamSpec
from typing import Sequence
from typing import TypeVar
from typing import Union


class Hole:
    def __init__(self, name: str = ""):
        self._name = name
        self._value = Underived()

    def __eq__(self, anything_is_equal_to_this):
        if self.is_unset:
            self._value = anything_is_equal_to_this
            return True
        return self._value == anything_is_equal_to_this

    @property
    def value(self):
        return self._value

    @property
    def is_set(self):
        return not self.is_unset

    @property
    def is_unset(self):
        return type(self.value) is Underived

    def __req__(self, anything_is_equal_to_this):
        if type(self.value) is Underived:
            self._value = anything_is_equal_to_this
            return True
        return self._value == anything_is_equal_to_this

    def __iter__(self):
        return repeat(Hole())

    def __add__(self, x):
        return Hole()

    def __radd__(self, x):
        return Hole()

    def __sub__(self, x):
        return Hole()

    def __rsub__(self, x):
        return Hole()

    def __mul__(self, x):
        return Hole()

    def __rmul__(self, x):
        return Hole()

    def __truediv__(self, x):
        return Hole()

    def __rtruediv__(self, x):
        return Hole()

    def __str__(self):
        return str(self.value)

    def __getitem__(self, key):
        return Hole(lambda x: self.unpack(x)["key"])


class PlaceholderHole(Hole): ...


class DerivableHole(Hole):
    def __init__(self, noisy=False):
        Hole.__init__(self)
        self._noisy = noisy  # todo: this is for debug and understanding only

    def __eq__(self, x):
        if self._noisy:
            print("Derived: ", x)
        return Hole.__eq__(self, x)

    def __req__(self, x):
        if self._noisy:
            print("Derived: ", x)
        return Hole.__eq__(self, x)


class Underived:
    def __init__(self):
        pass

    def __str__(self):
        return "UNDERIVED"


class IncompleteDerivationError(ValueError): ...


placeholder_hole = PlaceholderHole()
derivable_hole = DerivableHole()
noisy_derivable_hole = DerivableHole(noisy=True)

Parameters = ParamSpec("Parameters")
ReducedParameters = ParamSpec("ReducedParameters")
MissingParameters = ParamSpec("MissingParameters")
Return = TypeVar("Return")
FunctionToBeWrapped = Callable[Parameters, Return]
InnerFunction = Callable[
    Parameters, Callable[Parameters, Union[Return, Callable[MissingParameters, Return]]]
]


def infinite_generator(value):
    while True:
        yield value


class Unspecified:
    def __init__(self):
        pass

    def __str__(self):
        return "UNSPECIFIED"


unspecified_parameter = Unspecified()


# doing: replace REPLACEMENT OF HOLES with CALLING OF HOLES
def hole_aware(f: FunctionToBeWrapped) -> InnerFunction:
    @wraps(f)
    def inner(
        *args_possibly_with_placeholders: Parameters,
        **kwargs_possibly_with_placeholders: Parameters,
    ) -> Union[Return, Callable[MissingParameters, Return]]:
        is_with_placeholder = False
        for arg in chain(
            args_possibly_with_placeholders, kwargs_possibly_with_placeholders.values()
        ):
            if isinstance(arg, PlaceholderHole):
                is_with_placeholder = True
                break

        if not is_with_placeholder:
            return f(
                *args_possibly_with_placeholders, **kwargs_possibly_with_placeholders
            )

        @hole_aware
        def almost_f(
            *missing_args: MissingParameters,
            **missing_kwargs_or_overwrites: MissingParameters,
        ) -> Return:
            missing_args_supply = chain(iter(missing_args))
            complete_args = []
            for arg in args_possibly_with_placeholders:
                if isinstance(arg, PlaceholderHole):
                    candidate_arg_for_hole = next(missing_args_supply)

                    supplied_arg = candidate_arg_for_hole

                    complete_args.append(supplied_arg)
                else:
                    complete_args.append(arg)

            complete_kwargs = (
                kwargs_possibly_with_placeholders | missing_kwargs_or_overwrites
            )

            if _contains_placeholder_hole(complete_kwargs.values()):
                return almost_f(*complete_args, **complete_kwargs)

            return hole_aware(f)(*complete_args, **complete_kwargs)

        return almost_f

    inner._original_function = f
    return inner


def deriving_holes(f: FunctionToBeWrapped) -> InnerFunction:
    @wraps(f)
    def inner(*args, **kwargs):
        new_args = []
        new_kwargs = {}
        all_holes = []

        for arg in args:
            if type(arg) is DerivableHole:
                v = DerivableHole(noisy=arg._noisy)
                new_args.append(v)
                all_holes.append(v)
            else:
                new_args.append(arg)

        for k in kwargs:
            v = kwargs[k]
            if type(v) is DerivableHole:
                v = DerivableHole(noisy=v._noisy)
                new_kwargs[k] = v
                all_holes.append(v)
            else:
                new_kwargs[k] = v

        try:
            return f(*new_args, **new_kwargs)
        except:

            filled_args = []
            filled_kwargs = {}

            report_args = []
            report_kwargs = {}

            report_terms = {True: "Derived", False: "Underived", -1: "Given"}
            is_complete = True

            for arg in new_args:
                if type(arg) is DerivableHole:
                    report_args.append(report_terms[arg.is_set])
                    filled_args.append(arg.value)
                else:
                    report_args.append(report_terms[-1])
                    filled_args.append(arg)

            for k in new_kwargs:
                v = new_kwargs[k]
                if type(v) is DerivableHole:
                    report_kwargs[k] = report_terms[v.is_set]
                    filled_kwargs[k] = v.value
                else:
                    report_kwargs[k] = report_terms[-1]
                    filled_kwargs[k] = v

            if not all([h.is_set for h in all_holes]):
                raise IncompleteDerivationError(
                    f"Cannot fully derive values for {f.__name__} args:{str(report_args)} kwargs: {str(report_kwargs)}"
                )

            return f(*filled_args, **filled_kwargs)

    return inner


def _contains_placeholder_hole(seq: Sequence[Any]) -> bool:
    for item in seq:
        if type(item) is PlaceholderHole:
            return True
    return False


def flexible(f):
    # deriving_holes is postponed as it is very tedious to write
    @hole_aware
    @wraps(f)
    def inner(*args, **kwargs):
        return f(*args, **kwargs)

    return inner



from typing import Callable
from typing import Union

from jaxtyping import Array

from zephyr.building._template.array import array
from zephyr.building.initializers import Initializer
from zephyr.building.initializers import initializer_base
from zephyr.building.skeleton import Skeleton
from zephyr.functools.partial import hole_aware
from zephyr.functools.partial import placeholder_hole as _
from zephyr.project_typing import Shape


def validate(
    params: Union[Array, Skeleton],
    shape: Shape = None,
    initializer: Initializer = initializer_base,
    expression: Callable = lambda params: True,
) -> Union[Array, Skeleton]:
    if type(params) is Skeleton and shape is None:
        return params
    if shape is None and expression is None:
        return params
    if shape is None and expression is not None:
        expression(params)
        return params

    if type(params) is Skeleton:
        params == array(shape, initializer)
        return params
    else:
        # print(params.shape, shape, "<---")
        if not all([params.shape == shape]):
            raise ValueError(
                f"Incompatible shapes: shape of `params`: {params.shape} should be {shape} as specified."
            )
    return params  # not reachable, just for mypy
